"""Shared braille glyphs for inline histograms and sparklines."""

from __future__ import annotations

FILL_CHAR = "\u283f"
SPROUT_CHAR = "\u2807"
FULL_BLOCK_CHAR = "\u28ff"

__all__ = ["FILL_CHAR", "SPROUT_CHAR", "FULL_BLOCK_CHAR"]
